/*
 * Core OCR algorithm (first exact bitmap match)
 */
/*
 *  This is part of ypp-sc-tools, a set of third-party tools for assisting
 *  players of Yohoho Puzzle Pirates.
 * 
 *  Copyright (C) 2009 Ian Jackson <ijackson@chiark.greenend.org.uk>
 * 
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 * 
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 *  Yohoho and Puzzle Pirates are probably trademarks of Three Rings and
 *  are used without permission.  This program is not endorsed or
 *  sponsored by Three Rings.
 */

#include "ocr.h"
#include "convert.h"

typedef struct {
  Pixcol col;
  struct DatabaseNode *then;
} DatabaseLink;

typedef struct DatabaseNode {
  char *str;
  int nlinks, alinks;
  unsigned match:1, defined:1, endsword:1;
  DatabaseLink *links;
} DatabaseNode;

typedef struct {
  OcrReader *rd;
  OcrCellType ct;
  int w;
  Pixcol *cols;
  int x;
  unsigned ctxmap;
} FindCharArgs;
typedef struct {
  DatabaseNode *match;
  int rx;
} FindCharResults;

enum { ct_Lower, ct_Upper, ct_Word, ct_Digit };
static const char *context_names[]= {
  "Lower",  /*  bit 0, value 001 */
  "Upper",  /*  bit 1, value 002 */
  "Word",   /*  bit 2, value 004 */
  "Digit",  /*  bit 3, value 010 */
};
struct OcrCellTypeInfo {
  /* bitmaps of indices into context_names: */
  unsigned initial, nextword, midword;
  int space_spaces;
  const char *name;
  int (*findchar_select)(const FindCharArgs *fca,
			 const FindCharResults results[]);
};

#define NCONTEXTS (sizeof(context_names)/sizeof(context_names[0]))

struct OcrReader {
  int h;
  DatabaseNode contexts[NCONTEXTS];
  OcrResultGlyph *results;
  int aresults, nresults;
};

DEBUG_DEFINE_DEBUGF(ocr)

#define FGETSLINE (dbfile_getsline(lbuf,sizeof(lbuf),__FILE__,__LINE__))

static void cleardb_node(DatabaseNode *n) {
  int i;
  free(n->str); n->str=0;
  n->defined=n->match=n->endsword= 0;
  for (i=0; i<n->nlinks; i++)
    cleardb_node(n->links[i].then);
}

static void readdb1(OcrReader *rd, const char *which);

static void readdb(OcrReader *rd) {
  int ctxi;
  
  for (ctxi=0; ctxi<NCONTEXTS; ctxi++)
    cleardb_node(&rd->contexts[ctxi]);

  readdb1(rd, "master");
  readdb1(rd, "local");
}

static void readdb1(OcrReader *rd, const char *which) {
  int nchrs;
  DatabaseNode *current, *additional;
  char chrs[100];
  Pixcol cv;
  int j,ctxi;
  int h, endsword;
  char lbuf[100];

  char *dbfname= masprintf("%s/#%s-char%d#.txt",
			   get_vardir(), which, rd->h);
  
  if (!dbfile_open(dbfname))
    goto x;

  FGETSLINE;
  dbassert(!strcmp(lbuf,"# ypp-sc-tools pctb font v1"));

  dbassert( dbfile_scanf("%d", &h) == 1);
  dbassert(h==rd->h);

  for (;;) {
    FGETSLINE;
    if (!lbuf[0] || lbuf[0]=='#') continue;
    if (!strcmp(lbuf,".")) break;

    for (ctxi=0; ctxi<NCONTEXTS; ctxi++)
      if (!strcmp(lbuf,context_names[ctxi]))
	goto found_ctx;
    /* not found, just skip */
    for (;;) { FGETSLINE; if (!lbuf[0]) break; }
    continue;

  found_ctx:
    for (nchrs=0;;) {
      int c= fgetc(dbfile); sysassert(!ferror(dbfile)); dbassert(c!=EOF);
      if (c=='\n') break; /* forces no match */
      dbassert(nchrs<sizeof(chrs));
      chrs[nchrs++]= c;
    }
    endsword= 0;
    if (nchrs>0 && chrs[nchrs-1]==' ') {
      endsword= 1;
      nchrs--;
    }

    current= &rd->contexts[ctxi];
    for (;;) {
      FGETSLINE;
      if (!lbuf[0]) { dbassert(current != &rd->contexts[ctxi]); break; }
      char *ep;
      cv= strtoul(lbuf,&ep,16);  dbassert(!*ep);
      dbassert(!(cv & ~((1UL << rd->h)-1)));
      
      for (j=0; j<current->nlinks; j++)
	if (current->links[j].col == cv) {
	  current= current->links[j].then;
	  goto found_link;
	}

      additional= mmalloc(sizeof(*additional));
      additional->str= 0;
      additional->defined= 0;
      additional->match= 0;
      additional->endsword= 0;
      additional->nlinks= additional->alinks= 0;
      additional->links= 0;
      if (current->nlinks==current->alinks) {
	current->alinks++;
	current->alinks<<=1;
	current->links= mrealloc(current->links,
				 sizeof(*current->links) * current->alinks);
      }
      current->links[current->nlinks].col= cv;
      current->links[current->nlinks].then= additional;
      current->nlinks++;
      current= additional;

    found_link:;
    }

    if (!current->defined) {
      free(current->str);
      current->str= 0;
      current->defined= 1;
      current->match= 0;

      if (nchrs) {
	current->str= mmalloc(nchrs+1);
	memcpy(current->str, chrs, nchrs);
	current->str[nchrs]= 0;
	current->match= 1;
	current->endsword= endsword;
      }
    }
  }
 x:
  dbfile_close();
  free(dbfname);
}

static void cu_pr_ctxmap(FILE *resolver, unsigned ctxmap) {
  fprintf(resolver,"{");
  const char *spc="";
  int ctxi;
  for (ctxi=0; ctxi<NCONTEXTS; ctxi++) {
    if (!(ctxmap & (1u << ctxi))) continue;
    fprintf(resolver,"%s%s",spc,context_names[ctxi]);
    spc=" ";
  }
  fprintf(resolver,"}");
}

static void callout_unknown(OcrReader *rd, int w, Pixcol cols[],
			    int unk_l, int unk_r, unsigned unk_ctxmap) {
  int c,i, x,y;
  const OcrResultGlyph *s;
  const char *p;
  Pixcol pv;

  FILE *resolver= resolve_start();
  if (!resolver || !(o_flags & ff_editcharset))
    fatal("OCR failed - unrecognised characters or ligatures.\n"
	  "Character set database needs to be updated or augmented.\n"
	  "See README.charset.\n");
  
  fprintf(resolver,
	  "char\n"
	  "%d %d ",unk_l,unk_r);
  cu_pr_ctxmap(resolver,unk_ctxmap);
  for (i=0, s=rd->results; i<rd->nresults; i++, s++) {
    if (!strcmp(s->s," ")) continue;
    fprintf(resolver," %d %d ",s->l,s->r);
    cu_pr_ctxmap(resolver, 1u << s->ctxi);
    fprintf(resolver," ");
    for (p=s->s; (c= *p); p++) {
      if (c=='\\') fprintf(resolver,"\\%c",c);
      else if (c>=33 && c<=126) fputc(c,resolver);
      else fprintf(resolver,"\\x%02x",(unsigned char)c);
    }
  }
  fputc('\n',resolver);

  fprintf(resolver,
	  "/* XPM */\n"
	  "static char *t[] = {\n"
	  "/* columns rows colors chars-per-pixel */\n"
	  "\"%d %d 2 1\",\n"
	  "\"  c black\",\n"
	  "\"o c white\",\n",
	  w,rd->h);
  for (y=0, pv=1; y<rd->h; y++, pv<<=1) {
    fputc('"',resolver);
    for (x=0; x<w; x++)
      fputc(cols[x] & pv ? 'o' : ' ', resolver);
    fputs("\",\n",resolver);
  }
  fputs("};\n",resolver);

  resolve_finish();
  readdb(rd);
}

static void add_result(OcrReader *rd, const char *s, int l, int r, int ctxi) {
  if (rd->nresults >= rd->aresults) {
    rd->aresults++; rd->aresults<<=1;
    rd->results= mrealloc(rd->results, sizeof(*rd->results)*rd->aresults);
  }
  rd->results[rd->nresults].s= s;
  rd->results[rd->nresults].l= l;
  rd->results[rd->nresults].r= r;
  rd->results[rd->nresults].ctxi= ctxi;
  rd->nresults++;
}


static DatabaseNode *findchar_1ctx(const FindCharArgs *fca,
				   DatabaseNode *start, int *matchx_r) {
  DatabaseNode *current= start;
  DatabaseNode *bestmatch= 0;
  int i;
  int x= fca->x;

  for (;;) {
    debug_flush();
    debugf(" | x=%d",x);
    if (x > fca->w) break;
    Pixcol cv= fca->cols[x];
    debugf(" cv=%"PSPIXCOL(PRIx),cv);
    for (i=0; i<current->nlinks; i++)
      if (current->links[i].col == cv)
	goto found;
    /* not found */
    debugf(" ?");
    break;

  found:
    current= current->links[i].then;
    if (current->match) {
      debugf(" \"%s\"%s",current->str,current->endsword?"_":"");
      bestmatch= current;
      *matchx_r= x;
    } else {
      debugf(" ...");
    }

    x++;
  }
  return bestmatch;
}  

static DatabaseNode *findchar(const FindCharArgs *fca,
			      int *match_rx, int *match_rctxi) {
  FindCharResults results[NCONTEXTS];
  int ctxi, match=-1, nmatches=0;

  debugf("OCR  lx=%d ct_state=%x  ", fca->x, fca->ctxmap);
  for (ctxi=0; ctxi<NCONTEXTS; ctxi++) {
    results[ctxi].match= 0;
    if (!(fca->ctxmap & (1u << ctxi))) continue;
    debugf(" || %s",context_names[ctxi]);

    results[ctxi].match= findchar_1ctx(fca, &fca->rd->contexts[ctxi],
				       &results[ctxi].rx);
    if (!results[ctxi].match) continue;

    match= ctxi;
    nmatches++;
  }
  if (nmatches==1) {
    debugf(" unique");
  } else {
    debugf(" ambiguous");
    match= !fca->ct->findchar_select ? -1 :
      fca->ct->findchar_select(fca,results);
    debugf(" resolved %s", match<0 ? "<none>" : context_names[match]);
  }
  if (match<0)
    return 0;
  
  *match_rx= results[match].rx;
  if (match_rctxi) *match_rctxi= match;
  return results[match].match;
}

static int findchar_select_text(const FindCharArgs *fca,
				const FindCharResults results[]) {
  if (fca->ctxmap != 017) return -1;

  dbassert(! results[ct_Digit].match );
  if (results[ct_Word].match) return ct_Word;
  if (results[ct_Lower].rx > results[ct_Upper].rx) return ct_Lower;
  if (results[ct_Upper].rx > results[ct_Lower].rx) return ct_Upper;
  return -1;
}

const struct OcrCellTypeInfo ocr_celltype_number= {
  010,010,010,
  .space_spaces= 5,
  .name= "number",
  .findchar_select= 0
};
const struct OcrCellTypeInfo ocr_celltype_text= {
  .initial=  012, /* Digit|Upper */
  .nextword= 017, /* Digit|Upper|Lower|Word */
  .midword=  011, /* Digit|Lower */
  .space_spaces= 4,
  .name= "text",
  .findchar_select= findchar_select_text
};


const char *ocr_celltype_name(OcrCellType ct) { return ct->name; }

OcrResultGlyph *ocr(OcrReader *rd, OcrCellType ct, int w, Pixcol cols[]) {
  int nspaces;
  int x;

  FindCharArgs fca;
  fca.rd= rd;
  fca.ct= ct;
  fca.w= w;
  fca.cols= cols;
  fca.x= -1;

 restart:

  nspaces=- w;
  fca.ctxmap= ct->initial;
  rd->nresults=0;
  debugf("OCR h=%d w=%d",rd->h,w);
  for (x=0; x<w; x++) debugf(" %"PSPIXCOL(PRIx),cols[x]);
  debugf("\n");
  debug_flush();

  x=0;
  for (;;) {
    debug_flush();
    /* skip spaces */
    if (x>=w)
      break;

    if (!cols[x]) {
      nspaces++;
      x++;
      if (nspaces == ct->space_spaces) {
	debugf("OCR  x=%x nspaces=%d space\n",x,nspaces);
	fca.ctxmap= ct->nextword;
      }
      continue;
    }

    /* something here, so we need to add the spaces */
    if (nspaces >= ct->space_spaces)
      add_result(rd," ",x-nspaces,x+1,-1);
    nspaces=0;

    fca.x= x;

    int match_rx=-1;
    int match_ctxi=-1;
    DatabaseNode *match= findchar(&fca, &match_rx, &match_ctxi);
    
    if (match) {
      debugf(" || YES");
      add_result(rd, match->str, x, match_rx, match_ctxi);
      x= match_rx+1;
      if (match->match) fca.ctxmap= ct->midword;
      else debugf(" (empty)");
      if (match->endsword) {
	nspaces= ct->space_spaces;
	debugf("_");
	fca.ctxmap= ct->nextword;
      }
      debugf("\n");
    } else {
      int rx;
      debugf(" || UNKNOWN");
      for (rx=x; rx<w && cols[rx]; rx++);
      debugf(" x=%d ctxmap=%x %d..%d\n",x, fca.ctxmap, x,rx);
      debug_flush();
      callout_unknown(rd, w,cols, x,rx-1, fca.ctxmap);
      goto restart;
    }

  }
  add_result(rd, 0,-1,-1,-1);
  debugf("OCR  finished %d glyphs\n",rd->nresults);
  debug_flush();
  return rd->results;
}

OcrReader *ocr_init(int h) {
  OcrReader *rd;

  if (o_flags & ff_dict_fetch) {
    char *fetchfile= masprintf("char%d",h);
    progress("Updating %s...",fetchfile);
    fetch_with_rsync(fetchfile);
    free(fetchfile);
  }

  rd= mmalloc(sizeof(*rd));
  memset(rd,0,sizeof(*rd));
  rd->h= h;
  readdb(rd);
  return rd;
}
